/*
A GLUT program using OpenGL, GLUI, and GLSL.

It draws a little "particle system".

Orion Sky Lawlor, olawlor@acm.org, 2007/01/18 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

#include "osl/random.h"
#include "osl/random_vector.h"
#include "osl/random.cpp"

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"


/* Variables updated by the GUI */
float distance=1.5; /* origin-to-camera distance */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

float pointsize=1.0; /* size of points */
int draw_levels=7; /* levels of the IFS to draw */
int draw_outline=1; /* draw outlines for the matrices? */
int draw_stalks=0; /* draw stalks up to this level */

/* GUI representation of a matrix */
struct matrix_gui {
	vec3 trans; /* translation part */
	float scale; /* scale part */
	mat4 rot; /* rotation part */
	
	matrix_gui() :trans(0,0,0), scale(1.0), rot(1.0) {}
};

/* Matrices in our IFS */
std::vector<matrix_gui> matrices(4);

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 modelMat(1.0), viewMat(1.0);

/// Make a translation matrix by this offset 
mat4 makeTranslation(const vec3 &off) {
	mat4 m=mat4(1.0);
	m[3][0]=off[0]; m[3][1]=off[1]; m[3][2]=off[2]; 
	return m;
}
/// Make a scaling matrix by these values
mat4 makeScale(const vec3 &scl) {
	mat4 m=mat4(1.0);
	m[0][0]=scl[0]; m[1][1]=scl[1]; m[2][2]=scl[2]; 
	return m;
}
/// Scale by this offset
inline void glScalefv(const float *s) {
	glScalef(s[0],s[1],s[2]);
}
/// Translate by this offset
inline void glTranslatefv(const float *s) {
	glTranslatef(s[0],s[1],s[2]);
}
/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. */
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}


mat4 make_matrix(const matrix_gui &m) {
	mat4 t=makeTranslation(m.trans);
	mat4 s=makeScale(vec3(m.scale));
	return t*s*m.rot;
}

/******************* Particle System ************************/
oglProgramObject *dot_program=0; /* particle-drawing program */

class Particle {
public:
	Particle(const vec3 &pos_,const vec4 &color_) 
		:pos(pos_), color(color_)
	{
		
	}
	
	void draw(void) {
		glColor4fv(color);
		glVertex3fv(pos);
	}
	vec3 pos; /* position of particle in 3D space */
	vec4 color;
};

/* Draw a wireframe box, representing this matrix */
void drawBox(const mat4 &m) {
	glPushMatrix();
	glMultMatrixf(&m[0][0]);
	glTranslated(0.5,0.5,0.5);
	glutWireCube(1.0);
	glPopMatrix();
}

/* A set of animating Particles */
class ParticleSystem {
public:
	ParticleSystem();
	
	void draw(void) {
	/* Set up the matrices */
		if (draw_outline) {glColor4f(0.5,0.5,1.0,1);drawBox(mat4(1.0));}
		ifs_matrices.resize(0);
		for (unsigned int m=0;m<matrices.size();m++) {
			mat4 im=make_matrix(matrices[m]);
			ifs_matrices.push_back(im);
			if (draw_outline) {glColor4f(1,1,1,1);drawBox(im);}
		}

	/* Recursively generate particles */
		particles.resize(0,Particle(vec3(0.0f),vec4(0.0f)));
		glBegin(GL_LINES);
		fill_particles(mat4(1.0),0);
		glEnd();
	
	/* Draw the particles */
		dot_program->begin();
		glBegin(GL_POINTS);
		for (unsigned int i=0;i<particles.size();i++)
			particles[i].draw();
		glEnd();
		dot_program->end();
	}
	
	/* Recursive routine to fill "particles" array with IFS particle locations. */
	void fill_particles(const mat4 &parentMatrix,int levels) {
		vec3 loc=vec3(parentMatrix*vec4(0,0,0,1));
		particles.push_back(Particle(loc,vec4(loc.x,loc.y,loc.z,1)));
		if (levels<=draw_levels) { /* Recursively expand particles of our children */
			levels++;
			for (unsigned int i=0;i<ifs_matrices.size();i++)
				fill_particles(parentMatrix*ifs_matrices[i],levels);
			
			if (levels<=draw_stalks) 
			{ /* Draw lines through center of each coordinate system */
				glColor3fv(loc);
				glVertex3fv(loc);
				vec3 end=vec3(parentMatrix*vec4(0,1.0,0,1));
				glColor3fv(end);
				glVertex3fv(end);
			}
		}
	}
	
	/* Estimate the center of mass of our IFS, by looking at a few points */
	vec3 center_of_mass(void) {
		vec3 sum=vec3(0.0); int n=0;
		for (unsigned int i=0;i<particles.size();i+=100) {sum+=particles[i].pos; n++;}
		return sum/n;
	}
private:
	osl::Random48 r;
	std::vector<mat4> ifs_matrices;
	std::vector<Particle> particles;
};

ParticleSystem::ParticleSystem()
	:r(3)
{
	
}

ParticleSystem particles; /*<- silly global variable */

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	static double last_elapsed=0.0;
	glFinish(); double elapsed=oglTimer();
	// Clear the color and depth buffers
	//glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	glClearColor(0.0,0.0,0.0,0.0); /* dark gray */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		win_w/(float)win_h, /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	glTranslatef(0,0,-distance); /* push world origin down the *unrotated* Z axis */
	glMultMatrixf(&viewMat[0][0]); /* camera orientation: could also go under GL_MODELVIEW */
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMultMatrixf(&modelMat[0][0]);
	// Shift so the IFS's center of mass is right in the middle of the screen
	glTranslatefv(-particles.center_of_mass()); // -0.5,-0.5,-0.5);

/********** Pretty Particle Rendering ***************/
	/* Set up splat texture */
	static oglTexture *splat=0;
	if (splat==0) {
		/* Make a new splat texture: a central dark spot
		  with fuzzy edges.  This is a greyscale texture.
		*/
		const int w=32, h=32;
		unsigned char arr[w*h];
		for (int y=0;y<h;y++)
		for (int x=0;x<w;x++) {
			double dx=x/(double)w-0.5;
			double dy=y/(double)h-0.5;
			double r=sqrt(dx*dx+dy*dy);
			double out=1.0;
			double r_min=0.05, r_max=0.5;
			if (r<r_min) out=1.0;
			else if (r>r_max) out=0.0;
			else out=1.0-(r-r_min)/(r_max-r_min);
			arr[y*w+x]=(unsigned char)(255*out);
		}
		splat=new oglTexture(arr,w,h,
			oglTexture_linear, GL_LUMINANCE,GL_UNSIGNED_BYTE,
			GL_LUMINANCE8);
	}

	/* Set up programmable shader. */
	if (!dot_program) dot_program=oglProgramFromStrings(
"/* GLSL Vertex shader, to compute point color and size */\n"
"uniform vec4 cameraloc; // Location of camera\n"
"uniform float exposure; // overall exposure control--point diameter 1 unit from camera\n"
"varying vec4 color; // Material diffuse color\n"
"uniform float pixels_per_radian; \n"
"void main(void) {\n"
"	gl_Position=gl_ModelViewProjectionMatrix*gl_Vertex;\n"
"	color=gl_Color;\n"
"	vec4 position=gl_ModelViewMatrix*gl_Vertex;\n"
"	float d=distance(vec3(cameraloc),vec3(position));\n"
"	gl_PointSize=pixels_per_radian*0.01/d; //divide by z \n"
"	color.a *= clamp(gl_PointSize*gl_PointSize/4.0,0.0,1.0);\n"
"	gl_PointSize=max(2.0,gl_PointSize);\n"
"	\n"
"}\n"
,
"/* GLSL Fragment shader, to write actual pixels */\n"
"varying vec4 color;\n"
"uniform sampler2D splatTex;\n"
"void main(void) {\n"
"	gl_FragColor=color; \n"
"	gl_FragColor.a*=texture2D(splatTex,vec2(gl_TexCoord[0])).r;\n"
"}\n"
	);
	dot_program->begin();
	
	/* World location of camera: just back out all the matrix transforms */
	vec3 cameraloc=vec3(transpose(viewMat)*vec4(0,0,+distance,1));
	dot_program->set("cameraloc",cameraloc);
	dot_program->set("pixels_per_radian",pointsize*win_h/1.3);
	splat->bind();
	dot_program->set("splatTex",0);
	dot_program->end(); /* <- we'll re-enable the dot program when we're ready to draw dots... */
	
	/* Make gl_PointSize work in GLSL */
	glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
	
	/* GL Point Sprites (magically makes texture coordinates in fragment program) */
	glEnable(GL_POINT_SPRITE_ARB);
	glTexEnvi(GL_POINT_SPRITE_ARB,GL_COORD_REPLACE_ARB,GL_TRUE);
	
	glPointSize(50.0); /* set default point size--pixels across */
	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_ALPHA_TEST); /* too agressive--loses too many points */
	glDisable(GL_DEPTH_TEST); /* don't do Z buffer (screws up overlaps, esp. w/blending) */
	glEnable(GL_BLEND);
	
	/* Draw the points */
	particles.draw();
	
	/* Write the matrices to the "current.ifs" file */
	FILE *f=fopen("current.ifs","wb");
	if (f) {fwrite(&matrices[0],sizeof(matrix_gui),matrices.size(),f); fclose(f);}
	
	glFinish(); elapsed=oglTimer()-elapsed; if (elapsed<=1.0e-3) elapsed=1.0e-3;
	last_elapsed=elapsed;
	if (1) /* print frames per second */
	{
		/* overwrite all matrices with identity-- back to screen coordinates */
		glMatrixMode(GL_PROJECTION); glLoadIdentity();
		glMatrixMode(GL_MODELVIEW); glLoadIdentity(); 
		glColor4f(1,1,1,1);
		oglPrintf(oglVector3d(-0.9,-0.9,0.0),"%.2f ms/frame; %4.2f fps\n",
			1.0e3*elapsed,1.0/elapsed);
	}
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
	// Repeatedly ask to be redrawn
	glutPostRedisplay();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h-170;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");

	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	/* Set up view matrix so it's tilted forward a bit */
	vec3 up=normalize(vec3(0,1,0.2));
	viewMat[1]=vec4(up,0.0);
	viewMat[2]=vec4(vec3(up.x,-up.z,up.y),0.0);

#if 1
	/* Hardcode a little Sierpinski tetrahedron */
	matrices.resize(4);
	for (unsigned int i=0;i<matrices.size();i++) matrices[i].scale=0.5;
	matrices[0].trans=vec3(0,0,0);
	matrices[1].trans=vec3(0.5,0,0);
	matrices[2].trans=vec3(0.25,0,0.5);
	matrices[3].trans=vec3(0.25,0.5,0.25);
#elif 0
	/* Hardcode a flattened Menger sponge */
	float third=1.0/3;
	matrices.resize(8);
	for (unsigned int i=0;i<matrices.size();i++) matrices[i].scale=third;
	matrices[0].trans=vec3(0,0,0);
	matrices[1].trans=vec3(1*third,0,0);
	matrices[2].trans=vec3(2*third,0,0);
	matrices[3].trans=vec3(0,0,1*third);
	matrices[4].trans=vec3(2*third,0,1*third);
	matrices[5].trans=vec3(0,0,2*third);
	matrices[6].trans=vec3(1*third,0,2*third);
	matrices[7].trans=vec3(2*third,0,2*third);
#elif 0
	/* Hardcode a flat Sierpinski triangle */
	matrices.resize(3);
	for (unsigned int i=0;i<matrices.size();i++) matrices[i].scale=0.5;
	matrices[0].trans=vec3(0,0,0);
	matrices[1].trans=vec3(0.5,0,0);
	matrices[2].trans=vec3(0.25,0,0.5);
#else 
	/* Hardcode a tree's leaves */
	matrices.resize(3);
	for (unsigned int i=0;i<matrices.size();i++) matrices[i].scale=0.7;
	matrices[0].trans=vec3(0,1,0);
	matrices[1].trans=vec3(0,1,0);
	matrices[2].trans=vec3(0,1,0);
#endif

	if (argc>1) { /* read matrices from this IFS input file */
		matrices.resize(0);
		FILE *f=fopen(argv[1],"rb");
		if (f) {
			matrix_gui nm;
			while (fread(&nm,sizeof(nm),1,f)>0)
				matrices.push_back(nm); 
			fclose(f);
		}
	}
	
	if (1) {
		// Make some sliders (this is the only GLUI code you'll need!)
		GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
			GLUI_SUBWINDOW_TOP);
		/* This causes the main window to be redrawn when a GLUI control changes */
		glui->set_main_gfx_window( main_window );

		(new GLUI_Spinner( glui, "Distance:", &distance))->set_float_limits(0.01, 10.0);
		(new GLUI_Spinner( glui, "Pointsize:", &pointsize))->set_float_limits(0.03,10.0);
		new GLUI_Checkbox(glui,"Outline Matrices",&draw_outline);
		(new GLUI_Spinner( glui, "Levels:", &draw_levels))->set_int_limits(1,20);
		(new GLUI_Spinner( glui, "Stalks:", &draw_stalks))->set_int_limits(0,20);
		
		new GLUI_Column(glui);
		new GLUI_Rotation(glui,"Model:",&modelMat[0][0]);
		//new GLUI_Column(glui,false);
		//new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
		for (unsigned int i=0;i<matrices.size();i++) {
			char label[100]; sprintf(label,"Mat%d",i);
			new GLUI_Column(glui);
			new GLUI_Rotation(glui,label,&matrices[i].rot[0][0]);
			(new GLUI_Spinner(glui, "Scale:", &matrices[i].scale))->set_float_limits(0.0,1.3);
			(new GLUI_Spinner(glui, "Trans X:", &matrices[i].trans.x))->set_float_limits(-0.5,1.5);
			(new GLUI_Spinner(glui, "Trans Y:", &matrices[i].trans.y))->set_float_limits(-0.5,1.5);
			(new GLUI_Spinner(glui, "Trans Z:", &matrices[i].trans.z))->set_float_limits(-0.5,1.5);
		}
	}
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
